from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Optimization (minimization) functions """

import numpy as _np
import time as _time
import sys as _sys
import os as _os
import scipy.optimize as _spo

try:
    from scipy.optimize import Result as _optResult #for earlier scipy versions
except:
    from scipy.optimize import OptimizeResult as _optResult #for later scipy versions

from .customcg import fmax_cg


def minimize(fn,x0, method='cg', callback=None,
             tol=1e-10, maxiter=1000000, maxfev=None,
             stopval=None, jac=None):
    """
    Minimizes the function fn starting at x0.

    This is a gateway function to all other minimization routines within this
    module, providing a common interface to many different minimization methods
    (including and extending beyond those available from scipy.optimize).

    Parameters
    ----------
    fn : function
        The function to minimize.

    x0 : numpy array
        The starting point (argument to fn).

    method : string, optional
        Which minimization method to use.  Allowed values are:
        "simplex" : uses fmin_simplex
        "supersimplex" : uses fmin_supersimplex
        "customcg" : uses fmax_cg (custom CG method)
        "brute" : uses scipy.optimize.brute
        "basinhopping" : uses scipy.optimize.basinhopping with L-BFGS-B
        "swarm" : uses fmin_particle_swarm
        "evolve" : uses fmin_evolutionary (which uses DEAP)
        < methods available from scipy.optimize.minimize >

    callback : function, optional
        A callback function to be called in order to track optimizer progress.
        Should have signature: myCallback(x, f=None, accepted=None).  Note that
        create_obj_func_printer(...) function can be used to create a callback.

    tol : float, optional
        Tolerance value used for all types of tolerances available in a given method.

    maxiter : int, optional
        Maximum iterations.

    maxfev : int, optional
        Maximum function evaluations; used only when available, and defaults to maxiter.

    stopval : float, optional
        For basinhopping method only.  When f <= stopval then basinhopping outer loop
        will terminate.  Useful when a bound on the minimum is known.

    jac : function
        Jacobian function.

    Returns
    -------
    scipy.optimize.Result object
        Includes members 'x', 'fun', 'success', and 'message'.
    """

    if maxfev is None: maxfev = maxiter

    #Run Minimization Algorithm
    if method == 'simplex':
        solution = fmin_simplex(fn, x0, slide=1.0, tol=tol, maxiter=maxiter)

    elif method == 'supersimplex':
        solution = fmin_supersimplex(fn, x0, outer_tol=1.0, inner_tol=tol,
                                     max_outer_iter=100, min_inner_maxiter=100, max_inner_maxiter=maxiter)

    elif method == 'customcg':
        def fn_to_max(x):
            f = fn(x); return -f if f is not None else None

        if jac is not None:
            def dfdx_and_bdflag(x):
                j = -jac(x)
                bd = _np.zeros(len(j)) #never say fn is on boundary, since this is an analytic derivative
                return j, bd
        else:
            dfdx_and_bdflag = None

        solution = fmax_cg(fn_to_max, x0, maxiter, tol, dfdx_and_bdflag, None ) #Note: even though we maximize, return value is negated to conform to min routines

    elif method == 'brute':
        ranges = [ (0.0,1.0) ] * len(x0); Ns = 4  #params for 'brute' algorithm
        xmin, _ = _spo.brute(fn, ranges, (), Ns) #jac=jac
        #print "DEBUG: Brute fmin = ",fmin
        solution = _spo.minimize(fn,xmin,method="Nelder-Mead", options={}, tol=tol, callback = callback, jac=jac)

    elif method == 'basinhopping':
        def basin_callback(x, f, accept):
            if callback is not None: callback(x,f=f,accepted=accept)
            if stopval is not None and f <= stopval:
                return True #signals basinhopping to stop
            return False
        solution = _spo.basinhopping(fn, x0, niter=maxiter, T=2.0, stepsize=1.0,
                                               callback=basin_callback, minimizer_kwargs={'method': "L-BFGS-B", 'jac': jac})

        #DEBUG -- follow with Nelder Mead to make sure basinhopping found a minimum. (It seems to)
        #print "DEBUG: running Nelder-Mead:"
        #opts = { 'maxfev': maxiter, 'maxiter': maxiter }
        #solution = _spo.minimize(fn, solution.x, options=opts, method="Nelder-Mead", tol=1e-8, callback=callback)
        #print "DEBUG: done: best f = ",solution.fun

        solution.success = True #basinhopping doesn't seem to set this...

    elif method == 'swarm':
        solution = fmin_particle_swarm(fn, x0, tol, maxiter, popsize=1000) #, callback = callback)

    elif method == 'evolve':
        solution = fmin_evolutionary(fn, x0, num_generations = maxiter, num_individuals = 500)

#    elif method == 'homebrew':
#      solution = fmin_homebrew(fn, x0, maxiter)

    else:
        #Set options for different algorithms
        opts = {'maxiter': maxiter, 'disp': False }
        if method == "BFGS": opts['gtol'] = tol  #gradient norm tolerance
        elif method == "L-BFGS-B": opts['gtol'] = opts['ftol'] = tol  #gradient norm and fractional y-tolerance
        elif method == "Nelder-Mead": opts['maxfev'] = maxfev  #max fn evals (note: ftol and xtol can also be set)

        if method in ("BFGS","CG","Newton-CG","L-BFGS-B","TNC","SLSQP","dogleg","trust-ncg"): #use jacobian
            solution = _spo.minimize(fn, x0, options=opts, method=method, tol=tol, callback=callback, jac=jac)
        else:
            solution = _spo.minimize(fn, x0, options=opts, method=method, tol=tol, callback=callback)

    return solution



def fmin_supersimplex(fn, x0, outer_tol, inner_tol, max_outer_iter, min_inner_maxiter, max_inner_maxiter):
    """
    Minimize a function using repeated applications of the simplex algorithm.

    By varying the maximum number of iterations and repeatedly calling scipy's
    Nelder-Mead simplex optimization, this function performs as a robust (but
    slow) minimization.

    Parameters
    ----------
    fn : function
        The function to minimize.

    x0 : numpy array
        The starting point (argument to fn).

    outer_tol : float
        Tolerance of outer loop

    inner_tol : float
        Tolerance fo inner loop

    max_outer_iter : int
        Maximum number of outer-loop iterations

    min_inner_maxiter : int
        Minimum number of inner-loop iterations

    max_inner_maxiter : int
        Maxium number of outer-loop iterations

    Returns
    -------
    scipy.optimize.Result object
        Includes members 'x', 'fun', 'success', and 'message'.
    """
    f_init = fn(x0)
    f_final = f_init - 10*outer_tol #prime the loop
    x_start = x0

    i = 1
    cnt_at_same_maxiter = 1
    inner_maxiter = min_inner_maxiter

    while ( f_init-f_final > outer_tol or inner_maxiter < max_inner_maxiter) and i < max_outer_iter:
        if f_init-f_final <= outer_tol and inner_maxiter < max_inner_maxiter:
            inner_maxiter *= 10; cnt_at_same_maxiter = 1
        if cnt_at_same_maxiter > 10 and inner_maxiter > min_inner_maxiter:
            inner_maxiter /= 10; cnt_at_same_maxiter = 1
        f_init = f_final

        print(">>> fmin_supersimplex: outer iteration %d (inner_maxiter = %d)" % (i,inner_maxiter))
        i += 1; cnt_at_same_maxiter += 1

        opts = {'maxiter': inner_maxiter, 'maxfev': inner_maxiter, 'disp': False }
        inner_solution = _spo.minimize(fn,x_start,options=opts, method='Nelder-Mead',callback=None, tol=inner_tol)

        if not inner_solution.success:
            print("WARNING: fmin_supersimplex inner loop failed (tol=%g, maxiter=%d): %s" \
                % (inner_tol,inner_maxiter,inner_solution.message))

        f_final = inner_solution.fun
        x_start = inner_solution.x
        print(">>> fmin_supersimplex: outer iteration %d gives min = %f" % (i,f_final))

    solution = _optResult()
    solution.x = inner_solution.x
    solution.fun = inner_solution.fun
    if i < max_outer_iter:
        solution.success = True
    else:
        solution.success = False
        solution.message = "Maximum iterations exceeded"
    return solution



def fmin_simplex(fn, x0, slide=1.0, tol=1e-8, maxiter=1000):
    """
    Minimizes a function using a custom simplex implmentation.

    This was used primarily to check scipy's Nelder-Mead method
    and runs much slower, so there's not much reason for using
    this method.

    Parameters
    ----------
    fn : function
        The function to minimize.

    x0 : numpy array
        The starting point (argument to fn).

    slide : float, optional
        Affects initial simplex point locations

    tol : float, optional
        Relative tolerance as a convergence criterion.

    maxiter : int, optional
        Maximum iterations.

    Returns
    -------
    scipy.optimize.Result object
        Includes members 'x', 'fun', 'success', and 'message'.
    """

    # Setup intial values
    n = len(x0)
    f = _np.zeros(n+1)
    x = _np.zeros((n+1,n))

    x[0] = x0

    # Setup intial X range
    for i in range(1,n+1):
        x[i] = x0
        x[i,i-1] = x0[i-1] + slide

    # Setup intial functions based on x's just defined
    for i in range(n+1):
        f[i] = fn(x[i])

    # Main Loop operation, loops infinitly until break condition
    counter = 0
    while True:
        low = _np.argmin(f)
        high = _np.argmax(f)
        counter += 1

        # Compute Migration
        d = (-(n+1)*x[high]+sum(x))/n

        # Break if value is close
        if _np.sqrt(_np.dot(d,d)/n)<tol or counter == maxiter:
            solution = _optResult()
            solution.x = x[low]; solution.fun = f[low]
            if counter < maxiter:
                solution.success = True
            else:
                solution.success = False
                solution.message = "Maximum iterations exceeded"
            return solution

        newX = x[high] + 2.0*d
        newF = fn(newX)

        if newF <= f[low]:
            # Bad news, new value is lower than any other point => replace high point with new values
            x[high] = newX
            f[high] = newF
            newX = x[high] + d
            newF = fn(newX)

            # Check if need to expand
            if newF <= f[low]:
                x[high] = newX
                f[high] = newF

        else:
            # Good news, new value is higher than lowest point

            # Check if need to contract
            if newF <= f[high]:
                x[high] = newX
                f[high] = newF
            else:
                # Contraction
                newX = x[high] + 0.5*d
                newF = fn(newX)
                if newF <= f[high]:
                    x[high] = newX
                    f[high] = newF
                else:
                    for i in range(len(x)):
                        if i!=low:
                            x[i] = (x[i]-x[low])
                            f[i] = fn(x[i])



#TODO err_crit is never used?
def fmin_particle_swarm(f, x0, err_crit, iter_max, popsize=100, c1=2, c2=2):
    """
    A simple implementation of the Particle Swarm Optimization Algorithm.
    Pradeep Gowda 2009-03-16

    Parameters
    ----------
    f : function
        The function to minimize.

    x0 : numpy array
        The starting point (argument to fn).

    err_crit : float
        Critical error (i.e. tolerance).  Stops when error < err_crit.

    iter_max : int
        Maximum iterations.

    popsize : int, optional
        Population size.  Larger populations are better at finding the global
        optimum but make the algorithm take longer to run.

    c1 : float, optional
        Coefficient describing a particle's affinity for it's (local) maximum.

    c2 : float, optional
        Coefficient describing a particle's affinity for the best maximum any
        particle has seen (the current global max).

    Returns
    -------
    scipy.optimize.Result object
        Includes members 'x', 'fun', 'success', and 'message'.
    """
    dimensions = len(x0)
    LARGE = 1e10

    class Particle:
        pass

    #initialize the particles
    particles = []
    for i in range(popsize):
        p = Particle()
        p.params = x0 + 2 * (_np.random.random(dimensions) - 0.5)
        p.best = p.params[:]
        p.fitness = LARGE # large == bad fitness
        p.v = _np.zeros(dimensions)
        particles.append(p)

    # let the first particle be the global best
    gbest = particles[0]; ibest = 0
    # bDoLocalFitnessOpt = False

    #DEBUG
    #if False:
    #    import pickle as _pickle
    #    bestGaugeMx = _pickle.load(open("bestGaugeMx.debug"))
    #    lbfgsbGaugeMx = _pickle.load(open("lbfgsbGaugeMx.debug"))
    #    cgGaugeMx = _pickle.load(open("cgGaugeMx.debug"))
    #    initialGaugeMx = x0.reshape( (4,4) )
    #
    #    #DEBUG: dump line cut to plot
    #    nPts = 100
    #    print "DEBUG: best offsets = \n", bestGaugeMx - initialGaugeMx
    #    print "DEBUG: lbfgs offsets = \n", lbfgsbGaugeMx - initialGaugeMx
    #    print "DEBUG: cg offsets = \n", cgGaugeMx - initialGaugeMx
    #
    #    print "# DEBUG plot"
    #    #fDebug = open("x0ToBest.dat","w")
    #    #fDebug = open("x0ToLBFGS.dat","w")
    #    fDebug = open("x0ToCG.dat","w")
    #    #fDebug = open("LBFGSToBest.dat","w")
    #    #fDebug = open("CGToBest.dat","w")
    #    #fDebug = open("CGToLBFGS.dat","w")
    #
    #    for i in range(nPts+1):
    #        alpha = float(i) / nPts
    #        #matM = (1.0-alpha) * initialGaugeMx + alpha*bestGaugeMx
    #        #matM = (1.0-alpha) * initialGaugeMx + alpha*lbfgsbGaugeMx
    #        matM = (1.0-alpha) * initialGaugeMx + alpha*cgGaugeMx
    #        #matM = (1.0-alpha) * lbfgsbGaugeMx + alpha*bestGaugeMx
    #        #matM = (1.0-alpha) * cgGaugeMx + alpha*bestGaugeMx
    #        #matM = (1.0-alpha) * cgGaugeMx + alpha*lbfgsbGaugeMx
    #        print >> fDebug, "%g %g" % (alpha, f(matM.flatten()))
    #    exit()
    #
    #
    #    fDebug = open("lineDataFromX0.dat","w")
    #    min_offset = -1; max_offset = 1
    #    for i in range(nPts+1):
    #        offset = min_offset + float(i)/nPts * (max_offset-min_offset)
    #        print >> fDebug, "%g" % offset,
    #
    #        for k in range(len(x0)):
    #            x = x0.copy(); x[k] += offset
    #            try:
    #                print >> fDebug, " %g" % f(x),
    #            except:
    #                print >> fDebug, " nan",
    #        print >> fDebug, ""
    #
    #    print >> fDebug, "#END DEBUG plot"
    #    exit()
    #END DEBUG


    #err = 1e10
    for iter_num in range(iter_max):
        w = 1.0 #- i/iter_max

        #bDoLocalFitnessOpt = bool(iter_num > 20 and abs(lastBest-gbest.fitness) < 0.001 and iter_num % 10 == 0)
        # lastBest = gbest.fitness
        # minDistToBest = 1e10; minV = 1e10; maxV = 0 #DEBUG

        for (ip,p) in enumerate(particles):
            fitness = f(p.params)

            #if bDoLocalFitnessOpt:
            #    opts = {'maxiter': 100, 'maxfev': 100, 'disp': False }
            #    local_soln = _spo.minimize(f,p.params,options=opts, method='L-BFGS-B',callback=None, tol=1e-2)
            #    p.params = local_soln.x
            #    fitness = local_soln.fun

            if fitness < p.fitness:  #low 'fitness' is good b/c we're minimizing
                p.fitness = fitness
                p.best = p.params

            if fitness < gbest.fitness:
                gbest = p; ibest = ip

            v = w*p.v + c1 * _np.random.random() * (p.best - p.params) \
                    + c2 * _np.random.random() * (gbest.params - p.params)
            p.params = p.params + v
            for (i,pv) in enumerate(p.params):
                p.params[i] = ((pv+1) % 2) - 1 #periodic b/c on box between -1 and 1

            #from .. import tools as tools_
            #matM = p.params.reshape( (4,4) )  #DEBUG
            #minDistToBest = min(minDistToBest, _tools.frobeniusdist(
            #                                    bestGaugeMx,matM)) #DEBUG
            #minV = min( _np.linalg.norm(v), minV)
            #maxV = max( _np.linalg.norm(v), maxV)

        #print "DB: min diff from best = ", minDistToBest #DEBUG
        #print "DB: min,max v = ", (minV,maxV)


        #if False: #bDoLocalFitnessOpt:
        #    opts = {'maxiter': 100, 'maxfev': 100, 'disp': False }
        #    print "initial fun = ",gbest.fitness,
        #    local_soln = _spo.minimize(f,gbest.params,options=opts, method='L-BFGS-B',callback=None, tol=1e-5)
        #    gbest.params = local_soln.x
        #    gbest.fitness = local_soln.fun
        #    print "  final fun = ",gbest.fitness

        print("Iter %d: global best = %g (index %d)" % (iter_num, gbest.fitness, ibest))

        #if err < err_crit:  break  #TODO: stopping condition

    ## Uncomment to print particles
    #for p in particles:
    #    print 'params: %s, fitness: %s, best: %s' % (p.params, p.fitness, p.best)

    solution = _optResult()
    solution.x = gbest.params; solution.fun = gbest.fitness
    solution.success = True
#    if iter_num < maxiter:
#        solution.success = True
#    else:
#        solution.success = False
#        solution.message = "Maximum iterations exceeded"
    return solution



def fmin_evolutionary(f, x0, num_generations, num_individuals):
    """
    Minimize a function using an evolutionary algorithm.

    Uses python's deap package to perform an evolutionary
    algorithm to find a function's global minimum.

    Parameters
    ----------
    fn : function
        The function to minimize.

    x0 : numpy array
        The starting point (argument to fn).

    num_generations : int
        The number of generations to carry out. (similar to the number
        of iterations)

    num_individuals : int
        The number of individuals in each generation.  More individuals
        make finding the global optimum more likely, but take longer
        to run.


    Returns
    -------
    scipy.optimize.Result object
        Includes members 'x', 'fun', 'success', and 'message'.
    """

    import deap as _deap
    import deap.creator as _creator
    import deap.base as _base
    import deap.tools as _tools
    numParams = len(x0)

    # Create the individual class
    _creator.create("FitnessMin", _base.Fitness, weights=(-1.0,))
    _creator.create("Individual", list, fitness=_creator.FitnessMin)

    # Create initialization functions
    toolbox = _base.Toolbox()
    toolbox.register("random", _np.random.random)
    toolbox.register("individual", _tools.initRepeat, _creator.Individual,
                     toolbox.random, n=numParams) # fn to init an individual from a list of numParams random numbers
    toolbox.register("population", _tools.initRepeat, list, toolbox.individual) # fn to create a population (still need to specify n)

    # Create operation functions
    def evaluate(individual):
        return f( _np.array(individual) ),  #note: must return a tuple

    toolbox.register("mate", _tools.cxTwoPoint)
    toolbox.register("mutate", _tools.mutGaussian, mu=0, sigma=0.5, indpb=0.1)
    toolbox.register("select", _tools.selTournament, tournsize=3)
    toolbox.register("evaluate", evaluate)

    # Create the population
    pop = toolbox.population(n=num_individuals)

    # Evaluate the entire population
    fitnesses = list(map(toolbox.evaluate, pop))
    for ind, fit in zip(pop, fitnesses):
        ind.fitness.values = fit

    PROB_TO_CROSS = 0.5
    PROB_TO_MUTATE = 0.2

    # Initialize statistics
    stats = _tools.Statistics(key=lambda ind: ind.fitness.values)
    stats.register("avg", _np.mean)
    stats.register("std", _np.std)
    stats.register("min", _np.min)
    stats.register("max", _np.max)
    logbook = _tools.Logbook()

    #Run algorithm
    for g in range(num_generations):
        record = stats.compile(pop)
        logbook.record(gen=g, **record)
        print("Gen %d: %s" % (g,record))

        # Select the next generation individuals
        offspring = toolbox.select(pop, len(pop))
        # Clone the selected individuals
        offspring = list(map(toolbox.clone, offspring))

        # Apply crossover on the offspring
        for child1, child2 in zip(offspring[::2], offspring[1::2]):
            if _np.random.random() < PROB_TO_CROSS:
                toolbox.mate(child1, child2)
                del child1.fitness.values
                del child2.fitness.values

        # Apply mutation on the offspring
        for mutant in offspring:
            if _np.random.random() < PROB_TO_MUTATE:
                toolbox.mutate(mutant)
                del mutant.fitness.values

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # The population is entirely replaced by the offspring
        pop[:] = offspring

    #get best individual and return params
    indx_min_fitness = _np.argmin( [ ind.fitness.values[0] for ind in pop ] )
    best_params = _np.array(pop[indx_min_fitness])

    solution = _optResult()
    solution.x = best_params; solution.fun = pop[indx_min_fitness].fitness.values[0]
    solution.success = True
    return solution




#def fmin_homebrew(f, x0, maxiter):
#    """
#    Cooked up by Erik, this algorithm is similar to basinhopping but with some tweaks.
#
#    Parameters
#    ----------
#    fn : function
#        The function to minimize.
#
#    x0 : numpy array
#        The starting point (argument to fn).
#
#    maxiter : int
#        The maximum number of iterations.
#
#    Returns
#    -------
#    scipy.optimize.Result object
#        Includes members 'x', 'fun', 'success', and 'message'.
#    """
#
#    STEP = 0.01
#    MAX_STEPS = int(2.0 / STEP) # allow a change of at most 2.0
#    MAX_DIR_TRIES = 1000
#    T = 1.0
#
#    global_best_params = cur_x0 = x0
#    global_best = cur_f = f(x0)
#    N = len(x0)
#    trial_x0 = x0.copy()
#
#    for it in range(maxiter):
#
#        #Minimize using L-BFGS-B
#        opts = {'maxiter': maxiter, 'maxfev': maxiter, 'disp': False }
#        soln = _spo.minimize(f,trial_x0,options=opts, method='L-BFGS-B',callback=None, tol=1e-8)
#
#        # Update global best
#        if soln.fun < global_best:
#            global_best_params = soln.x
#            global_best = soln.fun
#
#        #check if we accept the new minimum
#        if soln.fun < cur_f or _np.random.random() < _np.exp( -(soln.fun - cur_f)/T ):
#            cur_x0 = soln.x; cur_f = soln.fun
#            print "Iter %d: f=%g accepted -- global best = %g" % (it, cur_f, global_best)
#        else:
#            print "Iter %d: f=%g declined" % (it, cur_f)
#
#        trial_x0 = None; numTries = 0
#        while trial_x0 is None and numTries < MAX_DIR_TRIES:
#            #choose a random direction
#            direction = _np.random.random( N )
#            numTries += 1
#
#            #print "DB: test dir %d" % numTries #DEBUG
#
#            #kick solution along random direction until the value of f starts to get smaller again (if it ever does)
#            #  (this indicates we've gone over a maximum along this direction)
#            last_f = cur_f
#            for i in range(1,MAX_STEPS):
#                test_x = cur_x0 + i*STEP * direction
#                test_f = f(test_x)
#                #print "DB: test step=%f: f=%f" % (i*STEP, test_f)
#                if test_f < last_f:
#                    trial_x0 = test_x
#                    print "Found new direction in %d tries, new f(x0) = %g" % (numTries,test_f)
#                    break
#                last_f = test_f
#
#        if trial_x0 is None:
#            raise ValueError("Maximum number of direction tries exceeded")
#
#    solution = _optResult()
#    solution.x = global_best_params; solution.fun = global_best
#    solution.success = True
##    if it < maxiter:
##        solution.success = True
##    else:
##        solution.success = False
##        solution.message = "Maximum iterations exceeded"
#    return solution



def create_obj_func_printer(objFunc, startTime=None):
    """
    Create a callback function that prints the value of an objective function.

    Parameters
    ----------
    objFunc : function
        The objective function to print.

    startTime : float (optional)
        A reference starting time to use when printing elapsed times. If None,
        then the system time when this function is called is used (which is 
        often what you want).

    Returns
    -------
    function
        A callback function which prints objFunc.
    """
    if startTime is None:
        startTime = _time.time() #for reference point of obj func printer

    def print_obj_func(x,f=None,accepted=None): # Just print the objective function value (used to monitor convergence in a callback)
        if f is not None and accepted is not None:
            print("%5ds %22.10f %s" % (_time.time()-startTime, f, 'accepted' if accepted else 'not accepted'))
        else:
            print("%5ds %22.10f" % (_time.time()-startTime, objFunc(x)))
    return print_obj_func



def _fwd_diff_jacobian(f, x0, eps=1e-10):
    y0 = f(x0)
    M = len(y0)
    N = len(x0)
    jac = _np.empty( (M,N), 'd' )

    for j in range(N):
        xj = x0.copy(); xj[j] += eps
        yj = f(xj)
        jac[:,j] = (yj-y0)/eps # df_dxj

    return jac

def check_jac(f, x0, jacToCheck, eps=1e-10, tol=1e-6, errType='rel'):
    """
    Checks a jacobian function using finite differences.

    Parameters
    ----------
    f : function
        The function to check.

    x0 : numpy array
        The point at which to check the jacobian.

    jacToCheck : function
        A function which should compute the jacobian of f at x0.

    eps : float, optional
        Epsilon to use in finite difference calculations of jacobian.

    tol : float, optional
        The allowd tolerance on the relative differene between the
        values of the finite difference and jacToCheck jacobians
        if errType == 'rel' or the absolute difference if errType == 'abs'.

    Returns
    -------
    errSum : float
        The total error between the jacobians.
    errs : list
        List of (row,col,err) tuples giving the error for each row and column.
    ffd_jac : numpy array
        The computed forward-finite-difference jacobian.
    """
    orig_stdout = _sys.stdout
    devnull = open(_os.devnull, 'w')

    try:
        _sys.stdout = devnull #redirect stdout to null during the many f(x) calls
        fd_jac = _fwd_diff_jacobian(f,x0,eps)
    finally:
        _sys.stdout = orig_stdout
        devnull.close()

    assert(jacToCheck.shape == fd_jac.shape)
    M,N = jacToCheck.shape

    errSum = 0; errs = []
    if errType == 'rel':
        for i in range(M):
            for j in range(N):
                err = _np.abs(fd_jac[i,j]-jacToCheck[i,j]) / (_np.abs(fd_jac[i,j])+1e-10)
                if err > tol: errs.append( (i,j,err) )
                errSum += err

    elif errType == 'abs':
        for i in range(M):
            for j in range(N):
                err = _np.abs(fd_jac[i,j]-jacToCheck[i,j])
                if err > tol: 
                    errs.append( (i,j,err) )
                    #print("DEBUG JAC CHECK (%d,%d): %g vs %g (diff = %g)" %
                    # (i,j,fd_jac[i,j],jacToCheck[i,j],fd_jac[i,j]-jacToCheck[i,j]))
                errSum += err

    errs.sort(key=lambda x: -x[2])

    if len(errs) > 0:
        maxabs = _np.max(_np.abs(jacToCheck))
        max_err_ratio = _np.max([ x[2]/maxabs for x in errs ])
        if max_err_ratio > 0.01: 
            print("Warning: jacobian_check has max err/jac_max = %g (jac_max = %g)" % (max_err_ratio,maxabs))

    return errSum, errs, fd_jac



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""
Variables for working with the 2-qubit gate set containing the gates
I*X(pi/2), I*Y(pi/2), X(pi/2)*I, Y(pi/2)*I, and CNOT.
"""

from . import gatestringconstruction as _strc
from . import gatesetconstruction as _setc
from . import spamspecconstruction as _spamc

description = "I*X(pi/2), I*Y(pi/2), X(pi/2)*I, Y(pi/2)*I, and CNOT gates"

gates = ['Gix','Giy','Gxi','Gyi','Gcnot']

fiducials16 = _strc.gatestring_list(
    [ (), ('Gix',), ('Giy',), ('Gix','Gix'),
      ('Gxi',), ('Gxi','Gix'), ('Gxi','Giy'), ('Gxi','Gix','Gix'),
      ('Gyi',), ('Gyi','Gix'), ('Gyi','Giy'), ('Gyi','Gix','Gix'),
      ('Gxi','Gxi'), ('Gxi','Gxi','Gix'), ('Gxi','Gxi','Giy'), ('Gxi','Gxi','Gix','Gix') ] )

fiducials36 = _strc.gatestring_list(
    [ (), ('Gix',), ('Giy',), ('Gix','Gix'), ('Gix','Gix','Gix'), ('Giy','Giy','Giy'),
      ('Gxi',), ('Gxi','Gix'), ('Gxi','Giy'), ('Gxi','Gix','Gix'), ('Gxi','Gix','Gix','Gix'), ('Gxi','Giy','Giy','Giy'),
      ('Gyi',), ('Gyi','Gix'), ('Gyi','Giy'), ('Gyi','Gix','Gix'), ('Gyi','Gix','Gix','Gix'), ('Gyi','Giy','Giy','Giy'),
      ('Gxi','Gxi'), ('Gxi','Gxi','Gix'), ('Gxi','Gxi','Giy'), ('Gxi','Gxi','Gix','Gix'), ('Gxi','Gxi','Gix','Gix','Gix'),
      ('Gxi','Gxi','Giy','Giy','Giy'), ('Gxi','Gxi','Gxi'), ('Gxi','Gxi','Gxi','Gix'), ('Gxi','Gxi','Gxi','Giy'),
      ('Gxi','Gxi','Gxi','Gix','Gix'), ('Gxi','Gxi','Gxi','Gix','Gix','Gix'), ('Gxi','Gxi','Gxi','Giy','Giy','Giy'),
      ('Gyi','Gyi','Gyi'), ('Gyi','Gyi','Gyi','Gix'), ('Gyi','Gyi','Gyi','Giy'), ('Gyi','Gyi','Gyi','Gix','Gix'),
      ('Gyi','Gyi','Gyi','Gix','Gix','Gix'), ('Gyi','Gyi','Gyi','Giy','Giy','Giy') ] )

fiducials = fiducials16
prepStrs = fiducials16

effectStrs = _strc.gatestring_list(
    [(), ('Gix',), ('Giy',), 
     ('Gix','Gix'), ('Gxi',), 
     ('Gyi',), ('Gxi','Gxi'), 
     ('Gxi','Gix'), ('Gxi','Giy'), 
     ('Gyi','Gix'), ('Gyi','Giy')] )

legacy_effectStrs = _strc.gatestring_list(
    [ (), ('Gix',), ('Giy',), ('Gxi',), ('Gyi',),
      ('Gix','Gxi'), ('Gxi','Giy'), ('Gyi','Gix'),
      ('Gyi','Giy'), ('Gxi','Gxi') ] )

germs = _strc.gatestring_list(
    [('Gxi',), ('Gyi',), ('Gix',), ('Giy',), ('Gcnot',),
     ('Gxi', 'Gyi'), ('Gix', 'Giy'), ('Giy', 'Gyi'), ('Gix', 'Gyi'),
     ('Gyi', 'Gcnot'), ('Giy', 'Gcnot'),
     ('Gxi', 'Gcnot', 'Gcnot'),
     ('Giy', 'Gxi', 'Gcnot'),
     ('Giy', 'Gcnot', 'Gyi'),
     ('Giy', 'Gyi', 'Gcnot'),
     ('Gix', 'Gxi', 'Gcnot'),
     ('Giy', 'Giy', 'Gcnot'),
     ('Giy', 'Gcnot', 'Gxi'),
     ('Gix', 'Giy', 'Gcnot'),
     ('Giy', 'Gxi', 'Gyi'),
     ('Gix', 'Giy', 'Gyi'),
     ('Gyi', 'Gyi', 'Gyi', 'Gxi'),
     ('Giy', 'Giy', 'Giy', 'Gix'),
     ('Gxi', 'Gyi', 'Gix', 'Giy'),
     ('Gcnot', 'Gix', 'Gyi', 'Gyi'),
     ('Gcnot', 'Gix', 'Gix', 'Gcnot'),
     ('Gxi', 'Gcnot', 'Gyi', 'Gyi'),
     ('Gyi', 'Gyi', 'Gyi', 'Gix'),
     ('Gix', 'Gix', 'Giy', 'Gcnot', 'Gcnot'),
     ('Gcnot', 'Giy', 'Giy', 'Gix', 'Giy'),
     ('Gyi', 'Gcnot', 'Gix', 'Giy', 'Gyi'),
     ('Giy', 'Gxi', 'Gcnot', 'Gxi', 'Gcnot'),
     ('Gyi', 'Gcnot', 'Gxi', 'Gcnot', 'Gxi'),
     ('Gyi', 'Gxi', 'Gyi', 'Gxi', 'Gxi', 'Gxi'),
     ('Gyi', 'Gxi', 'Gyi', 'Gyi', 'Gxi', 'Gxi'),
     ('Gyi', 'Gyi', 'Gyi', 'Gxi', 'Gyi', 'Gxi'),
     ('Gxi', 'Gxi', 'Gyi', 'Gxi', 'Gyi', 'Gyi'),
     ('Giy', 'Gix', 'Giy', 'Gix', 'Gix', 'Gix'),
     ('Giy', 'Gix', 'Giy', 'Giy', 'Gix', 'Gix'),
     ('Giy', 'Giy', 'Giy', 'Gix', 'Giy', 'Gix'),
     ('Gix', 'Gix', 'Giy', 'Gix', 'Giy', 'Giy'),
     ('Gcnot', 'Gyi', 'Giy', 'Gxi', 'Gix', 'Gcnot'),
     ('Gxi', 'Giy', 'Gxi', 'Gcnot', 'Gyi', 'Gix'),
     ('Gxi', 'Giy', 'Giy', 'Giy', 'Gcnot', 'Gxi'),
     ('Gcnot', 'Gxi', 'Gcnot', 'Gxi', 'Giy', 'Gix'),
     ('Gyi', 'Gix', 'Gyi', 'Gix', 'Gxi', 'Gxi'),
     ('Gix', 'Gcnot', 'Gxi', 'Gix', 'Gxi', 'Gcnot'),
     ('Gxi', 'Giy', 'Gyi', 'Gxi', 'Gcnot', 'Gcnot'),
     ('Gix', 'Gix', 'Giy', 'Gcnot', 'Giy', 'Gcnot', 'Gxi'),
     ('Giy', 'Gxi', 'Gcnot', 'Gix', 'Gix', 'Giy', 'Giy'),
     ('Gxi', 'Gcnot', 'Giy', 'Gyi', 'Gxi', 'Gix', 'Giy'),
     ('Gcnot', 'Gcnot', 'Gix', 'Gxi', 'Giy', 'Gxi', 'Gxi'),
     ('Gxi', 'Gix', 'Giy', 'Gyi', 'Gix', 'Gix', 'Gix'),
     ('Gxi', 'Gix', 'Gyi', 'Gix', 'Gyi', 'Giy', 'Gyi'),
     ('Gix', 'Gix', 'Gix', 'Gix', 'Gxi', 'Gxi', 'Gyi'),
     ('Giy', 'Gcnot', 'Gxi', 'Gyi', 'Gyi', 'Gcnot', 'Gix', 'Gcnot'),
     ('Gxi', 'Gyi', 'Gxi', 'Giy', 'Gxi', 'Giy', 'Gix', 'Giy'),
     ('Giy', 'Giy', 'Gyi', 'Gix', 'Gcnot', 'Gxi', 'Gyi', 'Gyi'),
     ('Gxi', 'Gix', 'Gcnot', 'Gyi', 'Gix', 'Gcnot', 'Gix', 'Giy'),
     ('Gix', 'Gxi', 'Gxi', 'Giy', 'Gxi', 'Gyi', 'Gix', 'Gcnot'),
     ('Gix', 'Gix', 'Gyi', 'Gxi', 'Giy', 'Gix', 'Gcnot', 'Gyi'),
     ('Gix', 'Giy', 'Gix', 'Gxi', 'Gix', 'Giy', 'Gxi', 'Gxi'),
     ('Giy', 'Gix', 'Gcnot', 'Gxi', 'Gcnot', 'Gxi', 'Gcnot', 'Gyi'),
     ('Gxi', 'Giy', 'Gix', 'Gix', 'Gxi', 'Giy', 'Gxi', 'Gcnot'),
     ('Gyi', 'Gyi', 'Gyi', 'Gyi', 'Gix', 'Giy', 'Gix', 'Gyi')
     ])

legacy_germs = _strc.gatestring_list(
    [('Giy',),
     ('Gxi',),
     ('Gyi',),
     ('Gcnot',),
     ('Gix', 'Gyi'),
     ('Giy', 'Gyi'),
     ('Giy', 'Gcnot'),
     ('Gyi', 'Gcnot'),
     ('Gix', 'Gix', 'Giy'),
     ('Gix', 'Gix', 'Gyi'),
     ('Gix', 'Giy', 'Giy'),
     ('Gix', 'Giy', 'Gyi'),
     ('Gix', 'Giy', 'Gcnot'),
     ('Gix', 'Gxi', 'Gcnot'),
     ('Gix', 'Gcnot', 'Giy'),
     ('Gix', 'Gcnot', 'Gyi'),
     ('Giy', 'Giy', 'Gxi'),
     ('Giy', 'Giy', 'Gcnot'),
     ('Giy', 'Gxi', 'Gyi'),
     ('Giy', 'Gxi', 'Gcnot'),
     ('Giy', 'Gyi', 'Gxi'),
     ('Giy', 'Gyi', 'Gcnot'),
     ('Giy', 'Gcnot', 'Gxi'),
     ('Giy', 'Gcnot', 'Gyi'),
     ('Gxi', 'Gxi', 'Gyi'),
     ('Gxi', 'Gxi', 'Gcnot'),
     ('Gxi', 'Gyi', 'Gyi'),
     ('Gxi', 'Gcnot', 'Gcnot'),
     ('Gcnot', 'Gix', 'Gyi', 'Gyi'),
     ('Gcnot', 'Gix', 'Gix', 'Gcnot'),
     ('Gxi', 'Giy', 'Gix', 'Giy'),
     ('Gxi', 'Gyi', 'Gix', 'Giy'),
     ('Gyi', 'Gyi', 'Gyi', 'Gix'),
     ('Gxi', 'Gcnot', 'Gyi', 'Gyi'),
     ('Gyi', 'Gcnot', 'Gix', 'Giy', 'Gyi'),
     ('Gix', 'Gix', 'Giy', 'Gcnot', 'Gcnot'),
     ('Giy', 'Gxi', 'Gcnot', 'Gxi', 'Gcnot'),
     ('Gyi', 'Gcnot', 'Gxi', 'Gcnot', 'Gxi'),
     ('Gcnot', 'Giy', 'Giy', 'Gix', 'Giy'),
     ('Giy', 'Gix', 'Gix', 'Gyi', 'Gxi'),
     ('Giy', 'Gix', 'Gyi', 'Giy', 'Giy'),
     ('Gcnot', 'Gxi', 'Gcnot', 'Gix', 'Gyi'),
     ('Gix', 'Gix', 'Gyi', 'Gxi', 'Gyi', 'Gix'),
     ('Gxi', 'Giy', 'Gyi', 'Gxi', 'Gcnot', 'Gcnot'),
     ('Gcnot', 'Gyi', 'Gcnot', 'Gxi', 'Gyi', 'Gyi'),
     ('Gxi', 'Giy', 'Gxi', 'Gcnot', 'Gyi', 'Gix'),
     ('Gcnot', 'Gyi', 'Giy', 'Gxi', 'Gix', 'Gcnot'),
     ('Gcnot', 'Gxi', 'Gcnot', 'Gxi', 'Giy', 'Gix'),
     ('Gxi', 'Gxi', 'Giy', 'Gix', 'Giy', 'Gix'),
     ('Gxi', 'Giy', 'Giy', 'Giy', 'Gcnot', 'Gxi'),
     ('Gyi', 'Gix', 'Gyi', 'Gix', 'Gxi', 'Gxi'),
     ('Gix', 'Gcnot', 'Gxi', 'Gix', 'Gxi', 'Gcnot'),
     ('Gxi', 'Gix', 'Giy', 'Gyi', 'Gix', 'Gix', 'Gix'),
     ('Gyi', 'Gxi', 'Gyi', 'Giy', 'Giy', 'Gxi', 'Gix'),
     ('Gcnot', 'Gcnot', 'Gix', 'Gxi', 'Giy', 'Gxi', 'Gxi'),
     ('Gxi', 'Gcnot', 'Giy', 'Gyi', 'Gxi', 'Gix', 'Giy'),
     ('Gxi', 'Gix', 'Gyi', 'Gix', 'Gyi', 'Giy', 'Gyi'),
     ('Gix', 'Gcnot', 'Giy', 'Gcnot', 'Gcnot', 'Gix', 'Gix'),
     ('Giy', 'Gxi', 'Gcnot', 'Gix', 'Gix', 'Giy', 'Giy'),
     ('Gix', 'Gix', 'Gix', 'Gix', 'Gxi', 'Gxi', 'Gyi'),
     ('Gix', 'Gix', 'Giy', 'Gcnot', 'Giy', 'Gcnot', 'Gxi'),
     ('Gxi', 'Giy', 'Gix', 'Gix', 'Gxi', 'Giy', 'Gxi', 'Gcnot'),
     ('Giy', 'Gix', 'Gcnot', 'Gxi', 'Gcnot', 'Gxi', 'Gcnot', 'Gyi'),
     ('Gix', 'Giy', 'Gix', 'Gxi', 'Gix', 'Giy', 'Gxi', 'Gxi'),
     ('Gix', 'Gix', 'Gyi', 'Gxi', 'Giy', 'Gix', 'Gcnot', 'Gyi'),
     ('Gxi', 'Gyi', 'Gxi', 'Giy', 'Gxi', 'Giy', 'Gix', 'Giy'),
     ('Giy', 'Giy', 'Gyi', 'Gix', 'Gcnot', 'Gxi', 'Gyi', 'Gyi'),
     ('Gix', 'Gxi', 'Gxi', 'Giy', 'Gxi', 'Gyi', 'Gix', 'Gcnot'),
     ('Gyi', 'Gyi', 'Gyi', 'Gyi', 'Gix', 'Giy', 'Gix', 'Gyi'),
     ('Gxi', 'Gix', 'Gcnot', 'Gyi', 'Gix', 'Gcnot', 'Gix', 'Giy'),
     ('Giy', 'Gcnot', 'Gxi', 'Gyi', 'Gyi', 'Gcnot', 'Gix', 'Gcnot')] )

#Construct the target gateset
gs_target = _setc.build_gateset(
    [4], [('Q0','Q1')],['Gix','Giy','Gxi','Gyi','Gcnot'],
    [ "I(Q0):X(pi/2,Q1)", "I(Q0):Y(pi/2,Q1)", "X(pi/2,Q0):I(Q1)", "Y(pi/2,Q0):I(Q1)", "CX(pi,Q0,Q1)" ],
    prepLabels=['rho0'], prepExpressions=["0"],
    effectLabels=['E0','E1','E2'], effectExpressions=["0","1","2"],
    spamdefs={'upup': ('rho0','E0'), 'updn': ('rho0','E1'),
              'dnup': ('rho0','E2'), 'dndn': ('rho0','remainder') }, basis="pp")


specs16x10 = _spamc.build_spam_specs(
    prepStrs=prepStrs,
    effectStrs=effectStrs,
    prep_labels=gs_target.get_prep_labels(),
    effect_labels=gs_target.get_effect_labels() )

specs16 = _spamc.build_spam_specs(
    fiducials16,
    prep_labels=gs_target.get_prep_labels(),
    effect_labels=gs_target.get_effect_labels() )

specs36 = _spamc.build_spam_specs(
    fiducials36,
    prep_labels=gs_target.get_prep_labels(),
    effect_labels=gs_target.get_effect_labels() )

specs = specs16x10 #use smallest specs set as "default"



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""
Routines for converting python objects to latex.  Parallel rountines as
HtmlUtil has for HTML conversion.
"""

import numpy as _np
import cmath
from .. import objects as _objs

#Define basestring in python3 so unicode
# strings can be tested for in python2 using
# python2's built-in basestring type.
# When removing __future__ imports, remove
# this and change basestring => str below.
try:  basestring
except NameError: basestring = str


def latex(x, brackets=False, precision=6,
          polarprecision=3, sciprecision=0):
    """
    Convert a numpy array, number, or string to latex.

    Parameters
    ----------
    x : anything
        Value to convert into latex.

    brackets : bool, optional
        Whether to include brackets in the output for array-type variables.

    precision : int, optional
        Rounding for normal numbers

    polarprecision : int, optional
        Rounding for polars

    sciprecision : int, optional
        Precision with which to el when precision
        requires use of scientific notation.


    Returns
    -------
    string
        latex string for x.
    """
    if isinstance(x,_np.ndarray) or \
       isinstance(x,_objs.Gate) or \
       isinstance(x,_objs.SPAMVec):
        d = 0
        for l in x.shape:
            if l > 1: d += 1
        x = _np.squeeze(x)
        if d == 0: return latex_value(x, precision=precision, polarprecision=polarprecision, sciprecision=sciprecision)
        if d == 1: return latex_vector(x, brackets=brackets, precision=precision, polarprecision=polarprecision, sciprecision=sciprecision)
        if d == 2: return latex_matrix(x, brackets=brackets, precision=precision, polarprecision=polarprecision, sciprecision=sciprecision)
        raise ValueError("I don't know how to render a rank %d numpy array as latex" % d)
    elif type(x) in (list,tuple):
        return latex_list(x, precision=precision, polarprecision=polarprecision, sciprecision=sciprecision)
    elif type(x) in (float,int,complex,_np.float64,_np.int64):
        return latex_value(x, precision=precision, polarprecision=polarprecision, sciprecision=sciprecision)
    elif isinstance(x,basestring):
        return latex_escaped(x)
    else:
        print("Warning: %s not specifically converted to latex" % str(type(x)))
        return str(x)


def latex_list(l, brackets=False, precision=6, polarprecision=3, sciprecision=0):
    """
    Convert a python list to latex tabular column.

    Parameters
    ----------
    l : list
        list to convert into latex.

    brackets : bool, optional
        Whether to include brackets in the output for array-type variables.

    precision : int, optional
        Rounding for normal numbers

    polarprecision : int, optional
        Rounding for polars

    sciprecision : int, optional
        Precision with which to el when precision
        requires use of scientific notation.


    Returns
    -------
    string
        latex string for l.
    """
    lines = [ ]
    for el in l:
        lines.append( latex(el, brackets, precision, polarprecision, sciprecision) )
    return "\\begin{tabular}{c}\n" + \
                " \\\\ \n".join(lines) + "\n \end{tabular}\n"


def latex_vector(v, brackets=False, precision=6,
                 polarprecision=3, sciprecision=0):
    """
    Convert a 1D numpy array to latex.

    Parameters
    ----------
    v : numpy array
        1D array to convert into latex.

    brackets : bool, optional
        Whether to include brackets in the output latex.

    precision : int, optional
        Rounding for normal numbers

    polarprecision : int, optional
        Rounding for polars

    sciprecision : int, optional
        Precision with which to el when precision
        requires use of scientific notation.


    Returns
    -------
    string
        latex string for v.
    """
    lines = [ ]
    for el in v:
        lines.append( latex_value(el, precision=precision,
                                  polarprecision=polarprecision,
                                  sciprecision=sciprecision) )
    if brackets:
        return "$ \\left(\\!\\!\\begin{array}{c}\n" + \
                " \\\\ \n".join(lines) + "\n \end{array}\\!\\!\\right) $\n"
    else:
        return "$ \\begin{array}{c}\n" + \
                " \\\\ \n".join(lines) + "\n \end{array} $\n"


def latex_matrix(m, fontsize=None, brackets=False, precision=6,
                 polarprecision=3, sciprecision=0):
    """
    Convert a 2D numpy array to latex.

    Parameters
    ----------
    m : numpy array
        2D array to convert into latex.

    fontsize : int, optional
        If not None, the fontsize.

    brackets : bool, optional
        Whether to include brackets in the output latex.

    precision : int, optional
        Rounding for normal numbers

    polarprecision : int, optional
        Rounding for polars

    sciprecision : int, optional
        Precision with which to el when precision
        requires use of scientific notation.


    Returns
    -------
    string
        latex string for m.
    """
    lines = [ ]; prefix = ""
    if fontsize is not None:
        prefix += "\\fontsize{%f}{%f}\selectfont " % (fontsize, fontsize*1.2)

    for r in range(m.shape[0]):
        lines.append( " & ".join(
                [latex_value(el, precision=precision,
                             polarprecision=polarprecision, 
                             sciprecision=sciprecision) for el in m[r,:] ] ) )

    if brackets:
        return prefix + "$ \\left(\\!\\!\\begin{array}{%s}\n" % ("c" * m.shape[1]) + \
        " \\\\ \n".join(lines) + "\n \end{array}\\!\\!\\right) $\n"
    else:
        return prefix + "$ \\begin{array}{%s}\n" % ("c" * m.shape[1]) + \
        " \\\\ \n".join(lines) + "\n \end{array} $\n"



def latex_value(el, precision=6, polarprecision=3,
                sciprecision=0, complexAsPolar=True):
    """
    Convert a floating point or complex value to latex.

    Parameters
    ----------
    el : float or complex
        Value to convert into latex.

    precision : int, optional
        Rounding for normal numbers

    polarprecision : int, optional
        Rounding for polars

    sciprecision : int, optional
        Precision with which to el when precision
        requires use of scientific notation.

    complexAsPolar : bool, optional
        Whether to output complex values in polar form.  If False, usual
        a+ib form is used.


    Returns
    -------
    string
        latex string for el.
    """
    # ROUND = digits to round values to
    TOL = 1e-9  #tolerance for printing zero values

    def render(x):
        if abs(x) < 5*10**(-(precision+1)):
            s = "%.*e" % (sciprecision,x)
        elif abs(x) < 1:
            s = "%.*f" % (precision,x)
        elif abs(x) <= 10**precision:
            s = "%.*f" % (precision-int(_np.log10(abs(x))),x)  #round to get precision+1 digits when x is > 1
        else:
            s = "%.*e" % (sciprecision,x)

        #Fix scientific notition
        p = s.split('e')
        if len(p) == 2:
            ex = str(int(p[1])) #exponent without extras (e.g. +04 => 4)
            s = p[0] + "\\e{" + ex + "}"

        #Strip superfluous endings
        if "." in s:
            while s.endswith("0"): s = s[:-1]
            if s.endswith("."): s = s[:-1]
        return s

    if isinstance(el,basestring):
        return el
    if type(el) in (int,_np.int64):
        return "%d" % el
    if el is None or _np.isnan(el): return "--"

    try:
        if abs(el.real) > TOL:
            if abs(el.imag) > TOL:
                if complexAsPolar:
                    r,phi = cmath.polar(el)
                    ex = ("i%.*f" % (polarprecision, phi)) if phi >= 0 \
                        else ("-i%.*f" % (polarprecision, -phi))
                    s = "%se^{%s}" % (render(r),ex)
                else:
                    s = "%s%s%si" % (render(el.real),'+' if el.imag > 0 else '-', render(abs(el.imag)))
            else:
                s = "%s" % render(el.real)
        else:
            if abs(el.imag) > TOL:
                s = "%si" % render(el.imag)
            else:
                s = "0"
    except:
        #try:
        #    if abs(el) > TOL: #throws error if el is not a number
        #        s = "%s" % render(el.real)
        #    else:
        #        s = "0"
        #except:
        s = str(el)

    return s


def latex_escaped(txt):
    """
    Escape txt so it is latex safe.

    Parameters
    ----------
    txt : string
        value to escape

    Returns
    -------
    string
    """
    ret = txt.replace("_","\_")
    return ret



from __future__ import division, print_function, absolute_import, unicode_literals
import numpy as _np

class ProtectedArray(object):
    """
    A numpy ndarray-like class that allows certain elements to
    be treated as read-only.
    """

    def __init__(self, input_array, indicesToProtect=None):
        self.base = input_array

        #Get protected indices, a specified as:
        self.indicesToProtect = []
        if indicesToProtect is not None:
            if not (isinstance(indicesToProtect, tuple)
                    or isinstance(indicesToProtect, list)):
                indicesToProtect = (indicesToProtect,)

            assert(len(indicesToProtect) <= len(self.base.shape))
            for ky,L in zip(indicesToProtect,self.base.shape):
                if isinstance( ky, slice ):
                    pindices = range(*ky.indices(L))
                elif isinstance( ky, int ):
                    i = ky+L if ky<0 else ky
                    if i < 0 or i > L:
                        raise IndexError("index (%d) is out of range." % ky)
                    pindices = (i,)
                elif isinstance( ky, list ):
                    pindices = ky
                else: raise TypeError("Invalid index type: %s" % type(ky))
                self.indicesToProtect.append(pindices)

        if len(self.indicesToProtect) == 0:
            self.indicesToProtect = None
        self.base.flags.writeable = True


    #Mimic array behavior
    def __pos__(self):        return self.base
    def __neg__(self):        return -self.base
    def __abs__(self):        return abs(self.base)
    def __add__(self,x):      return self.base + x
    def __radd__(self,x):     return x + self.base
    def __sub__(self,x):      return self.base - x
    def __rsub__(self,x):     return x - self.base
    def __mul__(self,x):      return self.base * x
    def __rmul__(self,x):     return x * self.base
    def __truediv__(self, x):  return self.base / x
    def __rtruediv__(self, x): return x / self.base
    def __floordiv__(self,x):  return self.base // x
    def __rfloordiv__(self,x): return x // self.base
    def __pow__(self,x):      return self.base ** x
    def __eq__(self,x):       return self.base == x
    def __len__(self):        return len(self.base)
    def __int__(self):        return int(self.base)
    def __long__(self):       return int(self.base)
    def __float__(self):      return float(self.base)
    def __complex__(self):    return complex(self.base)



    #Pickle plumbing
    def __reduce__(self):
        return (ProtectedArray, (_np.zeros(self.base.shape),), self.__dict__)

    def __setstate__(self, state):
        self.__dict__.update(state)


    #Access to underlying ndarray
    def __getattr__(self, attr):
        # set references to our memory as (entirely) read-only
        ret = getattr(self.__dict__['base'],attr)
        if isinstance(ret, _np.ndarray) and ret.base is self.base:
            ret.flags.writeable = False
        return ret

    def __getslice__(self, i, j):
        #For special cases when getslice is still called, e.g. A[:]
        return self.__getitem__(slice(i, j))

    def __getitem__( self, key ) :

        writeable = True

        #check if key matches/overlaps protected region
        if self.indicesToProtect is not None:
            new_indicesToProtect = []; nUnprotectedIndices = 0
            tup_key = key if isinstance( key, tuple ) else (key,)

            while len(tup_key) < len(self.base.shape):
                tup_key = tup_key + (slice(None,None,None),)

            for ky,pindices,L in zip(tup_key,self.indicesToProtect,self.base.shape):

                #Get requested indices
                if isinstance( ky, slice ):
                    indices = range(*ky.indices(L))

                    new_pindices = []
                    for ii,i in enumerate(indices):
                        if i in pindices:
                            new_pindices.append( ii ) #index of i within indices
                    new_pindices = sorted(list(set(new_pindices)))
                    new_indicesToProtect.append(new_pindices)

                    #tally how many indices in this dimension are unprotected
                    nTotalInDim = len(indices)
                    nUnprotectedInCurDim = (len(indices) - len(new_pindices))

                elif isinstance( ky, int ):
                    i = ky+L if ky<0 else ky
                    if i > L:
                        raise IndexError("The index (%d) is out of range." % ky)

                    nTotalInDim = 1
                    if i not in pindices: #single index that is unprotected => all unprotected
                        nUnprotectedInCurDim = 1 # a single unprotected index
                    else:
                        nUnprotectedInCurDim = 0

                else: raise TypeError("Invalid index type: %s" % type(ky))

                nUnprotectedIndices += nUnprotectedInCurDim

                #if there exists a single dimension with no protected indices, then
                # the whole array is writeable.
                if nTotalInDim == nUnprotectedInCurDim:
                    writeable = True
                    new_indicesToProtect = None
                    break

            else: #if we didn't break b/c of above block, which means each dim has
                  # at least one protected index

                #if there are no unprotected indices, then just set writeable == False
                if nUnprotectedIndices == 0:
                    writeable = False
                    new_indicesToProtect = None
                else:
                    #There is at least one writeable (unprotected) index in some dimension
                    # and at least one protected index in *every* dimension. We need to
                    # set indicesToProtect to describe what to protect
                    assert( len(new_indicesToProtect) > 0 ) #b/c otherwise another case would hold
                    writeable = True
                    new_indicesToProtect = tuple(new_indicesToProtect)

        else: # (if nothing is protected)
            writeable = True
            new_indicesToProtect = None

        ret = _np.ndarray.__getitem__(self.base,key)

        if not _np.isscalar(ret):
            ret = ProtectedArray(ret)
            ret.base.flags.writeable = writeable
            ret.indicesToProtect = new_indicesToProtect
            #print "   writeable = ",ret.flags.writeable
            #print "   new_toProtect = ",ret.indicesToProtect
            #print "<< END getitem"
        return ret


    def __setitem__( self, key, val ) :
        #print "In setitem with key = ", key, "val = ",val

        protectionViolation = [] # per dimension
        if self.indicesToProtect is not None:
            tup_key = key if isinstance( key, tuple ) else (key,)
            for ky,pindices,L in zip(tup_key,self.indicesToProtect,self.base.shape):

                #Get requested indices
                if isinstance( ky, slice ):
                    indices = range(*ky.indices(L))
                    if any(i in pindices for i in indices):
                        protectionViolation.append(True)
                    else: protectionViolation.append(False)

                elif isinstance( ky, int ):
                    i = ky+L if ky<0 else ky
                    if i > L:
                        raise IndexError("The index (%d) is out of range." % ky)
                    protectionViolation.append( i in pindices )

                else: raise TypeError("Invalid index type: %s" % type(ky))

            if all(protectionViolation): #assigns to a protected index in each dim
                raise ValueError("**assignment destination is read-only")
        return self.base.__setitem__(key,val)












#SCRATCH: TO REMOVE
#    def __reduce__(self):
#        """ Pickle plumbing. """
#        createFn, args, state = _np.ndarray.__reduce__(self)
#        new_state = state + (self.__dict__,)
#        return (createFn, args, new_state)
#
#    def __setstate__(self, state):
#        """ Pickle plumbing. """
#        _np.ndarray.__setstate__(self,state[0:-1])
#        self.__dict__.update(state[-1])

    #def __array_prepare__(self, out_arr, context=None):
    #    print "PREPARE: ", type(out_arr)
    #    return _np.asarray(out_arr)

    #def __array_finalize__(self, obj):
    #    if obj is None: return # let __new__ handle flags
    #    #Note: can't condition off .base since it's not set yet...
    #    #bNone = obj.base is None
    #    print "FINALIZE"
    #
    #    print "FINALIZE obj = ",type(obj)
    #    print "FINALIZE self = ",type(self)
    #
    #    #self.flags.writeable = False #protect by default (in case getitem misses anything)
    #    self.flags.writeable = True  #un-protect by default.  We'd like to be able to control
    #      #this better, but if we set to False, then operations like flatten() can't work at
    #      # all.  Seems to be some numpy inconsistencies in this implementation...
    #    self.indicesToProtect = None #default, in case getitem misses anything

    #def flatten(self, order='C'):
    #    ret = _np.ndarray.flatten(self, order)
    #    self.writeable = True
    #    print "FLATTENED"
    #    return ret

    #def __copy__(self):
    #    print "__COPY__!!"
    #    return ProtectedArray(_np.asarray(self),self.indicesToProtect)
    #
    #def copy(self):
    #    print "COPY!!"
    #    return ProtectedArray(_np.asarray(self),self.indicesToProtect)



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Construction routines for SPAM specifiers """

from ..tools import remove_duplicates as _remove_duplicates
from ..objects import spamspec as _ss


def build_spam_specs(fiducialGateStrings=None, prepStrs=None, effectStrs=None, prepSpecs=None, effectSpecs=None, prep_labels=('rho0',), effect_labels=('E0',)):
    """
    Computes rho and E specifiers based on optional arguments.  This function
      is used to generate the (prepSpecs,effectSpecs) tuple needed by many of the
      Core GST routines.

    Parameters
    ----------
    fiducialGateStrings : list of (tuples or GateStrings), optional
        Each tuple contains gate labels specifying a fiducial gate string, and it
        is *assumed* that the zeroth rhoVec and EVec are used with this string
        to form a prep-specifier and effect-specifier, respectively.
        e.g. [ (), ('Gx',), ('Gx','Gy') ]

    prepStrs : list of (tuples or GateStrings), optional
        Each tuple contains gate labels and it is *assumed* that the zeroth rhoVec
        is used with this string to form a prep-specifier.
        e.g. [ ('Gi',) , ('Gx','Gx') , ('Gx','Gi','Gy') ]

    effectStrs : list of (tuples or GateStrings), optional
        Each tuple contains gate labels and it is *assumed* that the zeroth EVec
        is used with this string to form a effect-specifier.
        e.g. [ ('Gi',) , ('Gi') , ('Gx','Gi','Gy') ]

    prepSpecs : list of tuples, optional
        Each tuple contains gate labels followed by an integer indexing a rhoVec.
        e.g. [ ('Gi',0) , ('Gx','Gx',0) , ('Gx','Gi','Gy',0) ]

    effectSpecs : list of tuples, optional
        Each tuple contains an integer EVec index followed by gate labels.
        e.g. [ (0,'Gi') , (1,'Gi') , (0,'Gx','Gi','Gy') ]

    prep_labels : tuple of strs, optional
        Labels to prepend to fiducial strings to create prepSpecs

    effect_labels : tuple of strs, optional
        Labels to append to fiducial strings to create effectSpecs

    Returns
    -------
      (prepSpecs, effectSpecs)
          each of the form of the optional parameters prepSpecs and effectSpecs above.
    """

    if prepSpecs is not None:
        if prepStrs is not None or fiducialGateStrings is not None:
            raise ValueError("Can only specify one of prepSpecs, prepStrs, or fiducialGateStrings")
    elif prepStrs is not None:
        prepSpecs = [ _ss.SpamSpec(rhoLbl,f) for f in prepStrs for rhoLbl in prep_labels ]
        if fiducialGateStrings is not None:
            raise ValueError("Can only specify one of prepSpecs, prepStrs, or fiducialGateStrings")
    elif fiducialGateStrings is not None:
        prepSpecs = [ _ss.SpamSpec(rhoLbl,f) for f in fiducialGateStrings for rhoLbl in prep_labels ]
    else:
        raise ValueError("Must specfiy one of prepSpecs, prepStrs, or fiducialGateStrings")


    if effectSpecs is not None:
        if effectStrs is not None or fiducialGateStrings is not None:
            raise ValueError("Can only specify one of effectSpecs, effectStrs, or fiducialGateStrings")
    elif effectStrs is not None:
        effectSpecs = [ _ss.SpamSpec(eLbl,f) for f in effectStrs for eLbl in effect_labels ]
        if fiducialGateStrings is not None:
            raise ValueError("Can only specify one of effectSpecs, effectStrs, or fiducialGateStrings")
    elif fiducialGateStrings is not None:
        effectSpecs = [ _ss.SpamSpec(eLbl,f) for f in fiducialGateStrings for eLbl in effect_labels ]
    else:
        raise ValueError("Must specfiy one of effectSpecs, effectStrs, or fiducialGateStrings")

    return prepSpecs, effectSpecs


def get_spam_strs(specs):
    """
    Get just the string portion of a pair of rho and E specifiers by
      stripping last element of prepSpecs and first element of effectSpecs
      to get prepStrs and effectStrs.  Duplicate strings are removed.

    Parameters
    ----------
    specs : tuple
        (prepSpecs, effectSpecs) to extract strings portions from.

    Returns
    -------
    prepStrs : list of strings
    effectStrs : list of strings
    """
    prepSpecs, effectSpecs = specs
    prepStrs = _remove_duplicates( [ prepSpec.str for prepSpec in prepSpecs ] )
    effectStrs   = _remove_duplicates( [ effectSpec.str for effectSpec in effectSpecs ] )
    return prepStrs, effectStrs



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Chi-squared and related functions """

import numpy as _np

def chi2(dataset, gateset, gateStrings=None,
         returnGradient=False, returnHessian=False,
         minProbClipForWeighting=1e-4, clipTo=None,
         useFreqWeightedChiSq=False, check=False,
         memLimit=None):
    """
    Computes the total chi^2 for a set of gate strings.

    The chi^2 test statistic obtained by summing up the
    contributions of a given set of gate strings or all
    the strings available in a dataset.  Optionally,
    the gradient and/or Hessian of chi^2 can be returned too.

    Parameters
    ----------
    dataset : DataSet
        The data used to specify frequencies and counts

    gateset : GateSet
        The gate set used to specify the probabilities and SPAM labels

    gateStrings : list of GateStrings or tuples, optional
        List of gate strings whose terms will be included in chi^2 sum.
        Default value (None) means "all strings in dataset".

    returnGradient, returnHessian : bool
        Whether to compute and return the gradient and/or Hessian of chi^2.

    minProbClipForWeighting : float, optional
        defines the clipping interval for the statistical weight (see chi2fn).

    clipTo : 2-tuple, optional
        (min,max) to clip probabilities to within GateSet probability
        computation routines (see GateSet.bulk_fill_probs)

    useFreqWeightedChiSq : bool, optional
        Whether or not frequencies (instead of probabilities) should be used
        in statistical weight factors.

    check : bool, optional
        If True, perform extra checks within code to verify correctness.  Used
        for testing, and runs much slower when True.

    memLimit : int, optional
        A rough memory limit in bytes which restricts the amount of intermediate
        values that are computed and stored.


    Returns
    -------
    chi2 : float
        chi^2 value, equal to the sum of chi^2 terms from all specified gate strings
    dchi2 : numpy array
        Only returned if returnGradient == True. The gradient vector of
        length nGatesetParams, the number of gateset parameters.
    d2chi2 : numpy array
        Only returned if returnHessian == True. The Hessian matrix of
        shape (nGatesetParams, nGatesetParams).
    """

    # Scratch work:
    # chi^2 = sum_i N_i*(p_i-f_i)^2 / p_i  (i over gatestrings & spam labels)
    # d(chi^2)/dx = sum_i N_i * [ 2(p_i-f_i)*dp_i/dx / p_i - (p_i-f_i)^2 / p_i^2 * dp_i/dx ]
    #             = sum_i N_i * (p_i-f_i) / p_i * [2 - (p_i-f_i)/p_i   ] * dp_i/dx
    #             = sum_i N_i * t_i * [2 - t_i ] * dp_i/dx     where t_i = (p_i-f_i) / p_i
    # d2(chi^2)/dydx = sum_i N_i * [ dt_i/dy * [2 - t_i ] * dp_i/dx - t_i * dt_i/dy * dp_i/dx + t_i * [2 - t_i] * d2p_i/dydx ]
    #                          where dt_i/dy = [ 1/p_i - (p_i-f_i) / p_i^2 ] * dp_i/dy
    if useFreqWeightedChiSq:
        raise ValueError("frequency weighted chi2 is not implemented yet.")

    spamLabels = gateset.get_spam_labels() #this list fixes the ordering of the spam labels
    spam_lbl_rows = { sl:i for (i,sl) in enumerate(spamLabels) }
    vec_gs_len = gateset.num_params()

    if gateStrings is None:
        gateStrings = list(dataset.keys())

    nSpamLabels = len(spamLabels)
    nGateStrings = len(gateStrings)

    evTree = gateset.bulk_evaltree(gateStrings)

    #Memory allocation
    ns = nSpamLabels; ng = nGateStrings
    ne = gateset.num_params(); gd = gateset.get_dimension()
    C = 1.0/1024.0**3

    #  Estimate & check persistent memory (from allocs directly below)
    persistentMem = 8* (ng*(1 + 2*ns)) # in bytes
    if returnGradient or returnHessian: persistentMem += 8*ng*ns*ne
    if returnHessian: persistentMem += 8*ng*ns*ne**2
    if memLimit is not None and memLimit < persistentMem:
        raise MemoryError("Chi2 Memory limit (%g GB) is " % (memLimit*C) +
                          "< memory required to hold final results (%g GB)"
                          % (persistentMem*C))

    #  Allocate peristent memory
    N      = _np.empty( nGateStrings )
    f      = _np.empty( (nSpamLabels, nGateStrings) )
    probs  = _np.empty( (nSpamLabels, nGateStrings) )

    if returnGradient or returnHessian:
        dprobs = _np.empty( (nSpamLabels, nGateStrings,vec_gs_len) )
    if returnHessian:
        hprobs = _np.empty( (nSpamLabels, nGateStrings,vec_gs_len,vec_gs_len) )

    #  Estimate & check intermediate memory
    #    - maybe make GateSet methods get intermediate estimates?
    intermedMem = 8*ng*gd**2 # ~ bulk_product
    if returnGradient: intermedMem += 8*ng*gd**2*ne # ~ bulk_dproduct
    if returnHessian: intermedMem += 8*ng*gd**2*ne**2 # ~ bulk_hproduct
    if memLimit is not None and memLimit < intermedMem:
        reductionFactor = float(intermedMem) / float(memLimit)
        maxEvalSubTreeSize = int(ng / reductionFactor)
    else:
        maxEvalSubTreeSize = None

    if maxEvalSubTreeSize is not None:
        evTree.split(maxEvalSubTreeSize, None)

    #DEBUG - no verbosity passed in to just leave commented out
    #if memLimit is not None:
    #    print "Chi2 Memory estimates: (%d spam labels," % ns + \
    #        "%d gate strings, %d gateset params, %d gate dim)" % (ng,ne,gd)
    #    print "Peristent: %g GB " % (persistentMem*C)
    #    print "Intermediate: %g GB " % (intermedMem*C)
    #    print "Limit: %g GB" % (memLimit*C)
    #    if maxEvalSubTreeSize is not None:
    #        print "Maximum eval sub-tree size = %d" % maxEvalSubTreeSize
    #        print "Chi2 mem limit has imposed a division of evaluation tree:"
    #  evTree.print_analysis()


    for (i,gateStr) in enumerate(gateStrings):
        N[i] = float(dataset[gateStr].total())
        for k,sl in enumerate(spamLabels):
            f[k,i] = dataset[gateStr].fraction(sl)


    if returnHessian:
        gateset.bulk_fill_hprobs(hprobs, spam_lbl_rows, evTree,
                                probs, dprobs, clipTo, check)
    elif returnGradient:
        gateset.bulk_fill_dprobs(dprobs, spam_lbl_rows, evTree,
                                probs, clipTo, check)
    else:
        gateset.bulk_fill_probs(probs, spam_lbl_rows, evTree,
                                clipTo, check)


    #cprobs = _np.clip(probs,minProbClipForWeighting,1-minProbClipForWeighting) #clipped probabilities (also clip derivs to 0?)
    cprobs = _np.clip(probs,minProbClipForWeighting,1e10) #effectively no upper bound
    chi2 = _np.sum( N[None,:] * ((probs - f)**2/cprobs), axis=(0,1) ) # Note (0,1) are all axes in this case
    #TODO: try to replace final N[...] multiplication with dot or einsum, or do summing sooner to reduce memory

    if returnGradient:
        t = ((probs - f)/cprobs)[:,:,None] # (iSpamLabel, iGateString, 0) = (K,M,1)
        dchi2 = N[None,:,None] * t * (2 - t) * dprobs  # (1,M,1) * (K,M,1) * (K,M,N)  (K=#spam, M=#strings, N=#vec_gs)
        dchi2 = _np.sum( dchi2, axis=(0,1) ) # sum over gate strings and spam labels => (N)

    if returnHessian:
        dprobs_p = dprobs[:,:,None,:] # (K,M,1,N)
        t = ((probs - f)/cprobs)[:,:,None,None] # (iSpamLabel, iGateString, 0,0) = (K,M,1,1)
        dt = ((1.0/cprobs - (probs-f)/cprobs**2)[:,:,None] * dprobs)[:,:,:,None] # (K,M,1) * (K,M,N) = (K,M,N) => (K,M,N,1)
        d2chi2 = N[None,:,None,None] * (dt * (2 - t) * dprobs_p - t * dt * dprobs_p + t * (2 - t) * hprobs)
        d2chi2 = _np.sum( d2chi2, axis=(0,1) ) # sum over gate strings and spam labels => (N1,N2)
        # (1,M,1,1) * ( (K,M,N1,1) * (K,M,1,1) * (K,M,1,N2) + (K,M,1,1) * (K,M,N1,1) * (K,M,1,N2) + (K,M,1,1) * (K,M,1,1) * (K,M,N1,N2) )

    if returnGradient:
        return (chi2, dchi2, d2chi2) if returnHessian else (chi2, dchi2)
    else:
        return (chi2, d2chi2) if returnHessian else chi2



#def _oldTotalChiSquared( dataset, gateset, gateStrings=None, useFreqWeightedChiSq=False,
#                     minProbClipForWeighting=1e-4):
#    """
#    Return summed chi^2 values for given gateStrings or all gatestrings
#    in the given dataset (if gateStrings is None).
#    If useFreqWeightedChiSq is True, then use the dataset frequencies instead
#    of the (correct) gateset probabilities in the expression for "Chi^2".
#    """
#
#    if gateStrings is None:
#        return sum( [ gate_string_chi2( gatestring, dataset, gateset, useFreqWeightedChiSq,
#                                       minProbClipForWeighting) for gatestring in dataset] )
#        #return sum( [ chi2fn_2outcome( dsRow.total(),
#        #                     gateset.pr('plus', gatestring),
#        #                     dsRow.fraction('plus'),
#        #                     minProbClipForWeighting ) for gatestring,dsRow in dataset.iteritems() ] )
#    else:
#        return sum( [ gate_string_chi2( gatestring, dataset, gateset, useFreqWeightedChiSq,
#                                       minProbClipForWeighting) for gatestring in gateStrings] )
#

def gate_string_chi2( gatestring, dataset, gateset, useFreqWeightedChiSq=False,
                     minProbClipForWeighting=1e-4):
    """
    Computes the chi-squared term for a single gate string.

    The sum of the chi-squared terms for each of the possible
    outcomes (the SPAM labels in gateset).

    Parameters
    ----------
    gatestring : GateString or tuple
        The gate string to compute chi-squared for.

    dataset : Dataset
        The object from which frequencies are extracted.

    gateset : GateSet
        The object from which probabilities and SPAM labels are extracted.

    useFreqWeightedChiSq : bool, optional
        Whether or not frequencies (instead of probabilities) should be used
        in statistical weight factors.

    minProbClipForWeighting : float, optional
        If useFreqWeightedChiSq == False, defines the clipping interval
        for the statistical weight (see chi2fn).

    Returns
    -------
    float
        The sum of chi^2 terms for corresponding to gatestring
        (one term per SPAM label).
    """

    chiSqFn  = chi2fn_wfreqs if useFreqWeightedChiSq else chi2fn
    rowData  = dataset[gatestring]
    clip = minProbClipForWeighting

    N = rowData.total()
    p = gateset.probs(gatestring)
    return sum( [ chiSqFn(N, p[sl], rowData[sl]/N, clip) for sl in gateset.get_spam_labels() ] )


def chi2fn_2outcome( N, p, f, minProbClipForWeighting=1e-4 ):
    """
    Computes chi^2 for a 2-outcome measurement.

    The chi-squared function for a 2-outcome measurement using
    a clipped probability for the statistical weighting.

    Parameters
    ----------
    N : float or numpy array
        Number of samples.

    p : float or numpy array
        Probability of 1st outcome (typically computed).

    f : float or numpy array
        Frequency of 1st outcome (typically observed).

    minProbClipForWeighting : float, optional
        Defines clipping interval (see return value).

    Returns
    -------
    float or numpy array
        N(p-f)^2 / (cp(1-cp)),
        where cp is the value of p clipped to the interval
        (minProbClipForWeighting, 1-minProbClipForWeighting)
    """
    cp = _np.clip(p,minProbClipForWeighting,1-minProbClipForWeighting)
    return N*(p-f)**2/(cp*(1-cp))


def chi2fn_2outcome_wfreqs( N, p, f, minProbClipForWeighting=1e-4 ):
    """
    Computes chi^2 for a 2-outcome measurement using frequency-weighting.

    The chi-squared function for a 2-outcome measurement using
    the observed frequency in the statistical weight.

    Parameters
    ----------
    N : float or numpy array
        Number of samples.

    p : float or numpy array
        Probability of 1st outcome (typically computed).

    f : float or numpy array
        Frequency of 1st outcome (typically observed).

    minProbClipForWeighting : float, optional
        unused but present to keep the same function
        signature as chi2fn_2outcome.

    Returns
    -------
    float or numpy array
        N(p-f)^2 / (f*(1-f*)),
        where f* = (f*N+1)/N+2 is the frequency value used in the
        statistical weighting (prevents divide by zero errors)
    """
    f1 = (f*N+1)/(N+2)
    return N*(p-f)**2/(f1*(1-f1))


def chi2fn( N, p, f, minProbClipForWeighting=1e-4 ):
    """
    Computes the chi^2 term corresponding to a single outcome.

    The chi-squared term for a single outcome of a multi-outcome
    measurement using a clipped probability for the statistical
    weighting.

    Parameters
    ----------
    N : float or numpy array
        Number of samples.

    p : float or numpy array
        Probability of 1st outcome (typically computed).

    f : float or numpy array
        Frequency of 1st outcome (typically observed).

    minProbClipForWeighting : float, optional
        Defines clipping interval (see return value).

    Returns
    -------
    float or numpy array
        N(p-f)^2 / cp ,
        where cp is the value of p clipped to the interval
        (minProbClipForWeighting, 1-minProbClipForWeighting)
    """
    #cp = _np.clip(p,minProbClipForWeighting,1-minProbClipForWeighting)
    cp = _np.clip(p,minProbClipForWeighting,1e10) #effectively no upper bound
    return N*(p-f)**2/cp


def chi2fn_wfreqs( N, p, f, minProbClipForWeighting=1e-4 ):
    """
    Computes the frequency-weighed chi^2 term corresponding to a single outcome.

    The chi-squared term for a single outcome of a multi-outcome
    measurement using the observed frequency in the statistical weight.

    Parameters
    ----------
    N : float or numpy array
        Number of samples.

    p : float or numpy array
        Probability of 1st outcome (typically computed).

    f : float or numpy array
        Frequency of 1st outcome (typically observed).

    minProbClipForWeighting : float, optional
        unused but present to keep the same function
        signature as chi2fn.

    Returns
    -------
    float or numpy array
        N(p-f)^2 / f*,
        where f* = (f*N+1)/N+2 is the frequency value used in the
        statistical weighting (prevents divide by zero errors)
    """
    f1 = (f*N+1)/(N+2)
    return N*(p-f)**2/f1
